/* OpenICF3 isim */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINE  (1024)
#define MAXWORD  (32)

char DR[16][MAXLINE];
char label[256][MAXWORD];
unsigned long labeladr[256];
char cmd[32][MAXWORD];
int lp;
int xy,vcmd;

void delsp(char *p) {
    int i,j;

	for(i=j=0; p[i]!='\0' && i<MAXLINE ; i++)
		if(p[i] != ' ' && p[i] != '\n') p[j++] = p[i];
	if(i == MAXLINE) {
        printf("delsp() over\n");
		exit(-1);
    }
	p[j] = '\0';
}

void SetDR(char *p) {
	char *top;
	int eflag,num,i;

	eflag = 0;
    if(p[3]>='0' && p[3]<='9') {
        if(p[4] == '=') {
			top = &p[5];
			num = (p[2]-'0')*10+(p[3]-'0');
		} else eflag = 1;
	} else {
		if(p[3] == '=') {
			top = &p[4];
			num = p[2]-'0';
		} else eflag = 1;
	}
    if(eflag == 1) {
	    printf("$R error: '=' not found.\n");
		exit(-1);
	}
    if(num < 0 || num > 15) {
		printf("$R error: 0< or >15\n");
		exit(-1);
	}
    for(i=0 ; top[i] != '\0'; i++)
		DR[num][i] = top[i];
    DR[num][i] = '\0';
}

int isLabel(char *p,unsigned long adr) {
	int i;

	for(i=0 ; p[i] != ':' && p[i]!='\0' && i<MAXLINE ; i++)
	    label[lp][i] = p[i];
    if(p[i] == ':') {
		labeladr[lp] = adr;
		label[lp++][i] = '\0';
        if(lp > 255) {
		    printf("too many labels\n");
            exit(-1);
		}
        return 0;
	}
	return 1;
}

int StrCmp(char *p,const char *q) {
	int i;

    for(i=0 ; i<MAXWORD && p[i] != '\0' ; i++)
		if(p[i] != q[i] && q[i] != '?') return 0;
	if(i==MAXWORD) {
        printf("StrCmp Error\n");
		exit(-1);
	}
	if(q[i] != '\0') return 0;
	vcmd = 1;
	return 1;
}

int StrMch(char *p,const char *q) {
	int i;

    for(i=0 ; i<MAXWORD && q[i] != '\0' ; i++)
		if(p[i] != q[i] && q[i]!='?') return 0;
	if(i==MAXWORD) {
        printf("StrCmp2 Error\n");
		exit(-1);
	}
	vcmd = 1;
	return 1;
}

unsigned long branch(char *p,int adr) {
	char l[MAXWORD];
	int i,j;
	int count;
	int jmp;
	unsigned long ret;

	for(i=0,j=0 ; p[i]!='\0' && p[i] !=')' ; i++) l[j++] = p[i];
	l[j] = '\0';
    count = 0;
	for(i=0 ; i<lp ; i++) {
        if(StrCmp(l,label[i])) {
              jmp = (int)labeladr[i];
			  count++;
		}
	}
	if(count > 1) {
		printf("label conflict for branch.\n");
		exit(-1);
	}
	if(count == 0) {
		printf("not found label for branch.\n");
		exit(-1);
	}
	jmp = jmp - adr -2;
	if(jmp > 128) {
		printf("range error for branch.\n");
		exit(-1);
	}
	if( jmp < -129) {
		printf("too small address for branch.\n");
        exit(-1);
	}
    if(jmp < 0) jmp = 256+jmp;
	xy += 2;
	ret = (unsigned long) jmp;
	return ret;
}

unsigned long branchab(char *p) {
	char l[MAXWORD];
	int i,j;
	int count;
	unsigned long jmp;

	for(i=2,j=0 ; p[i]!='\0' && p[i] !=')' ; i++) l[j++] = p[i];
	l[j] = '\0';
    count = 0;
	for(i=0 ; i<lp ; i++) {
        if(StrCmp(l,label[i])) {
              jmp = labeladr[i];
			  count++;
		}
	}
	if(count > 1) {
		printf("label conflict for B command.\n");
		exit(-1);
	}
	if(count == 0) {
		printf("label not found for B command. %s\n",l);
		exit(-1);
	}
	if(jmp > 511) {
		printf("label too large for B command.\n");
		exit(-1);
	}
	xy += 2;
	return jmp;
}

int iasm_main(const char* asmfile) {
    int i,j,c;
    FILE *fp,*fw;
	int cp;
	char buf[MAXLINE];
	int num;
	unsigned long adr,code,debug;

	for(i=0 ; i<16 ; i++) {
		DR[i][0] = '0';
		DR[i][1] = '\0';
	}

    if((fw=fopen("bin","w"))==NULL) {
		printf("Can't open bin file\n");
		exit(-1);
	}

    /* PASS1 */
    if((fp=fopen(asmfile,"r"))==NULL) {
		printf("Can't open file\n");
		exit(-1);
	}
	adr = lp = 0;
    while( !feof(fp) ) {
		c = 0;
        i = 0;
		do {
		    c=fgetc(fp);
			if(c==EOF) break;
			if(c!='\r' && c!='\n') {
			    buf[i++]=c;
			} else {
			    if(i>0) break;
			}
		} while(!feof(fp) && i<MAXLINE);

		if(i == MAXLINE) {
			printf("line buffer over\n");
			exit(-1);
		}
		buf[i] = '\0';
        if(buf[0] == '#') continue;
		delsp(buf);
        if(buf[0] == '$' && buf[1] == 'R') {
			SetDR(buf);
			continue;
		}
		adr += isLabel(buf,adr);
    }
    fclose(fp);

	for(i=0 ; i<16 ; i++) fprintf(fw,"%s\n",DR[i]);
	
    /* PASS 2 */
    if((fp=fopen(asmfile,"r"))==NULL) {
		printf("Can't open file\n");
		exit(-1);
	}
    adr = 0;
    while( !feof(fp) ) {
	    c = 0;
        i = 0;
		do {
		    c=fgetc(fp);
			if(c==EOF) break;
			if(c!='\r' && c!='\n') {
			    buf[i++]=c;
			} else {
			    if(i>0) break;
			}
		} while(!feof(fp) && i<MAXLINE);
		
		if(i == MAXLINE) {
			printf("line buffer over\n");
			exit(-1);
		}
		buf[i] = '\0';
        if(buf[0] == -1) continue;
		if(buf[0]=='#' || buf[0]=='%' || buf[0]=='$') continue;
		c=0;
		for(i=0 ; buf[i] != '\0' ; i++) if(buf[i] == ':') c=1;
        if(c == 1) continue;
		delsp(buf);
		xy = 0;
		code = debug = 0;
        cp = 0;
		for(i=0,j=0 ; buf[i] != '\0' ; i++) {
			if(buf[i] == ';') {
				cmd[cp++][j] = '\0';
				j = 0;
			} else {
				cmd[cp][j++] = buf[i];
				if(j>MAXWORD-2) {
					printf("mnemonic too long.\n");
                    exit(-1);
				}
			}
		}
		cmd[cp++][j] = '\0';
        for(i=0 ; i<cp ; i++) {
			vcmd = 0;
            if(StrCmp(cmd[i],""))           {}
            if(StrCmp(cmd[i],"NOP"))        {}
			if(StrCmp(cmd[i],"I=160"))      {code |=0x10000000;}
			if(StrCmp(cmd[i],"I=1024"))     {code |=0x20000000;}
			if(StrMch(cmd[i],"I<=")) {
                for(num=0,j=3 ; cmd[i][j]>='0' && cmd[i][j]<='9' ; j++) 
					num = num*10 + (cmd[i][j]-'0');
				code |= 0x30000000 | (num-1);
				xy += 2;
            }
			if(StrCmp(cmd[i],"J=256"))      {code |=0x40000000;}
            if(StrCmp(cmd[i],"J=512"))      {code |=0x50000000;}
			if(StrCmp(cmd[i],"J=1024"))     {code |=0x60000000;}
            if(StrMch(cmd[i],"B("))         {code |=0x70000000|branchab(cmd[i]);}
            if(StrMch(cmd[i],"BZF("))       {code |=0x80000000|branch(&cmd[i][4],adr);}
			if(StrMch(cmd[i],"BC("))        {code |=0x90000000|branch(&cmd[i][3],adr);}
			if(StrMch(cmd[i],"BI("))        {code |=0xa0000000|branch(&cmd[i][3],adr);}
			if(StrMch(cmd[i],"BJ("))        {code |=0xb0000000|branch(&cmd[i][3],adr);}
			if(StrMch(cmd[i],"BCF("))       {code |=0xc0000000|branch(&cmd[i][4],adr);}
			if(StrMch(cmd[i],"CALL("))      {code |=0xd0000000|branch(&cmd[i][5],adr);}
			if(StrCmp(cmd[i],"RET"))        {code |=0xe0000000;}
			if(StrMch(cmd[i],"END("))       {code |=0xf0000000|(cmd[i][4]-'0');xy+=2;}
			if(StrCmp(cmd[i],"ADDL=0"))     {}
			if(StrCmp(cmd[i],"ADDL=A"))     {code |=0x04000000;}
			if(StrCmp(cmd[i],"ADDL=DIVA"))  {code |=0x08000000;}
			if(StrCmp(cmd[i],"ADDR=0"))     {}
			if(StrCmp(cmd[i],"ADDR=B"))     {code |=0x01000000;}
			if(StrCmp(cmd[i],"ADDR=C"))     {code |=0x02000000;}
			if(StrCmp(cmd[i],"ADDR=D"))     {code |=0x03000000;}
			if(StrCmp(cmd[i],"SUB"))        {code |=0x00800000;}
			if(StrCmp(cmd[i],"TESTD"))      {code |=0x00800000;}
			if(StrCmp(cmd[i],"INV"))        {code |=0x00400000;}
			if(StrCmp(cmd[i],"SEL(X)"))     {code |=0x00200000;}
			if(StrCmp(cmd[i],"CMP"))        {code |=0x00100000;}
			if(StrCmp(cmd[i],"XCOUNT"))     {code |=0x00100000;}
			if(StrCmp(cmd[i],"MUL"))        {code |=0x00080000;}
			if(StrMch(cmd[i],"R?=A"))       {code |=0x00040000|(cmd[i][1]-'0');xy++;}
            if(StrMch(cmd[i],"R1?=A" ))     {code |=0x00040000|(cmd[i][2]-'0'+10);xy++;}
            if(StrCmp(cmd[i],"A=SEL1"))     {code |=0x00010000;}
			if(StrCmp(cmd[i],"A=SEL2"))     {code |=0x00020000;}
			if(StrCmp(cmd[i],"A=SEL3"))     {code |=0x00030000;}
			if(StrCmp(cmd[i],"B=SEL1"))     {code |=0x00004000;}
			if(StrCmp(cmd[i],"B=SEL2"))     {code |=0x00008000;}
			if(StrCmp(cmd[i],"B=SEL3"))     {code |=0x0000C000;}
			if(StrCmp(cmd[i],"C=SEL1"))     {code |=0x00001000;}
			if(StrCmp(cmd[i],"C=SEL2"))     {code |=0x00002000;}
			if(StrCmp(cmd[i],"C=SEL3"))     {code |=0x00003000;}
			if(StrCmp(cmd[i],"D=SEL1"))     {code |=0x00000400;}
			if(StrCmp(cmd[i],"D=SEL2"))     {code |=0x00000800;}
			if(StrCmp(cmd[i],"D=SEL3"))     {code |=0x00000C00;}
			if(StrCmp(cmd[i],"MMUL"))       {code |=0x00000200;}
			if(StrCmp(cmd[i],"MRES"))       {code |=0x00000100;}
			if(StrCmp(cmd[i],"X=?"))        {code |= (cmd[i][2]-'0')<<4;}
			if(StrCmp(cmd[i],"X=1?"))       {code |= (cmd[i][3]-'0'+10)<<4;}
            if(StrCmp(cmd[i],"@A"))      debug |=0x80000000;
            if(StrCmp(cmd[i],"@B"))      debug |=0x40000000;
            if(StrCmp(cmd[i],"@C"))      debug |=0x20000000;
            if(StrCmp(cmd[i],"@D"))      debug |=0x10000000;
            if(StrCmp(cmd[i],"@M0A"))    debug |=0x08000000;
            if(StrCmp(cmd[i],"@M0B"))    debug |=0x04000000;
            if(StrCmp(cmd[i],"@M1A"))    debug |=0x02000000;
            if(StrCmp(cmd[i],"@M1B"))    debug |=0x01000000;
            if(StrCmp(cmd[i],"@CF"))     debug |=0x00800000;
            if(StrCmp(cmd[i],"@ZF"))     debug |=0x00400000;
            if(StrCmp(cmd[i],"@I"))      debug |=0x00200000;
            if(StrCmp(cmd[i],"@J"))      debug |=0x00100000;
            if(StrCmp(cmd[i],"@PC"))     debug |=0x00080000;
            if(StrCmp(cmd[i],"@RETADR")) debug |=0x00040000;
            if(StrCmp(cmd[i],"@ENDC"))   debug |=0x00020000;
            if(StrCmp(cmd[i],"@STOP"))   debug |=0x00010000;
            if(StrCmp(cmd[i],"@R0"))     debug |=0x00008000;
            if(StrCmp(cmd[i],"@R1"))     debug |=0x00004000;
            if(StrCmp(cmd[i],"@R2"))     debug |=0x00002000;
            if(StrCmp(cmd[i],"@R3"))     debug |=0x00001000;
            if(StrCmp(cmd[i],"@R4"))     debug |=0x00000800;
            if(StrCmp(cmd[i],"@R5"))     debug |=0x00000400;
            if(StrCmp(cmd[i],"@R6"))     debug |=0x00000200;
            if(StrCmp(cmd[i],"@R7"))     debug |=0x00000100;
            if(StrCmp(cmd[i],"@R8"))     debug |=0x00000080;
            if(StrCmp(cmd[i],"@R9"))     debug |=0x00000040;
			if(StrCmp(cmd[i],"@R10"))    debug |=0x00000020;
			if(StrCmp(cmd[i],"@R11"))    debug |=0x00000010;
			if(StrCmp(cmd[i],"@R12"))    debug |=0x00000008;
			if(StrCmp(cmd[i],"@R13"))    debug |=0x00000004;
			if(StrCmp(cmd[i],"@R14"))    debug |=0x00000002;
			if(StrCmp(cmd[i],"@R15"))    debug |=0x00000001;
     		if(vcmd == 0) {
	    		printf("Undefine NM : #%s#\n",cmd[i]);
		    	exit(-1);
			}
		}
		if(xy > 2) {
			printf("bit 24-31 conflict\n");
			exit(-1);
		}
        for(j=7 ; j>=0 ; j--) fprintf(fw,"%lX",(code>>(j*4))&0xf);
        fprintf(fw," ");
        for(j=7 ; j>=0 ; j--) fprintf(fw,"%lX",(debug>>(j*4))&0xf);
        fprintf(fw,"\n");

		adr++;
    }
    fclose(fw);
	return 0;
}
