/* OpenICF3 isim ver1.2 2016/7/16     */
/* x86: XP/Vista/7/8.1/10             */
/* x64: Vista/7/8.1/10                */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define RANGE_CHECK

const int PLEN = 512;

#ifdef __EMSCRIPTEN__
const int ALUSIZE = 32;
const int LASTBIT = ALUSIZE*32-1;
const int ALUBIT  = ALUSIZE*32;
typedef struct {
	unsigned long n[ALUSIZE];
    int carry;
} Integer;
#else
const int ALUSIZE = 16;
const int LASTBIT = ALUSIZE*64-1;
const int ALUBIT  = ALUSIZE*64;
typedef struct {
	unsigned long long n[ALUSIZE];
    int carry;
} Integer;
#endif

unsigned int time;
unsigned long PCODE[PLEN];
unsigned long PDEBUG[PLEN];
static FILE *fresult = NULL;

unsigned long long hex(char a) {
    if(a>='0' && a<='9') return a-'0';
	if(a>='A' && a<='F') return 10+a-'A';
	if(a>='a' && a<='f') return 10+a-'a';
	printf("Bad letter in atoI()\n");
    exit(-1);
}

void atoI(char*s, Integer* x) {
	int len,i,j,k;
	len = (int)strlen(s);
    if(len > ALUSIZE*16) {
		printf("too long %d bit in atoI()\n",ALUBIT);
		exit(-1);
	}
	for(i=0 ; i<ALUSIZE ; i++) x->n[i] = 0;
    for(i=0,j=0,k=0 ; i<len ; i++) {
	    x->n[j] |= hex(s[len-1-i])<<((k++)*4);
#ifdef __EMSCRIPTEN__
		k %= 8;
#else
		k %= 16;
#endif
		if(k==0) j++;
	}
}

int testbit(Integer* N,int p) {
#ifdef RANGE_CHECK
    if(p<0 || p>LASTBIT) {
	    printf("range error in testbit()\n");
		printf("p = %d\n",p);
		exit(-1);
	}
#endif
#ifdef __EMSCRIPTEN__
    return 1&((N->n[p/32])>>(p%32));
#else
    return 1&((N->n[p/64])>>(p%64));
#endif
}

void setbit(Integer* N,int p) {
#ifdef RANGE_CHECK
    if(p<0 || p>LASTBIT) {
	    printf("range error in setbit()\n");
		printf("p = %d\n",p);
		exit(-1);
	}
#endif
#ifdef __EMSCRIPTEN__
	N->n[p/32] |= (unsigned long)1 << (p%32);
#else
	N->n[p/64] |= (unsigned long long)1 << (p%64);
#endif
}

void i_copy(Integer* z,Integer* x) {
	for(int i=0 ; i<ALUSIZE ; i++) z->n[i] = x->n[i];
}

void i_zero(Integer* z) {
	for(int i=0 ; i<ALUSIZE ; i++) z->n[i] = 0;
}

void i_lshift(Integer *z) {
#ifdef __EMSCRIPTEN__
    z->carry = z->n[ALUSIZE-1] >> 31;
	for(int i=ALUSIZE-1 ; i>0 ; i--) z->n[i] = (z->n[i]<<1) | (z->n[i-1]>>31);
#else
    z->carry = z->n[ALUSIZE-1] >> 63;
	for(int i=ALUSIZE-1 ; i>0 ; i--) z->n[i] = (z->n[i]<<1) | (z->n[i-1]>>63);
#endif
    z->n[0] <<= 1;
}

void i_rshift(Integer *z) {
#ifdef __EMSCRIPTEN__
	for(int i=0 ; i<ALUSIZE-1 ; i++) z->n[i] = (z->n[i]>>1) | (z->n[i+1]<<31);
#else
	for(int i=0 ; i<ALUSIZE-1 ; i++) z->n[i] = (z->n[i]>>1) | (z->n[i+1]<<63);
#endif
	z->n[ALUSIZE-1] >>= 1;
}

void i_or(Integer *z, Integer *x, Integer *y) {
	for(int i=0 ; i<ALUSIZE ; i++) z->n[i] = (x->n[i]) | (y->n[i]);
}

#ifdef __EMSCRIPTEN__
void i_add(Integer *z,Integer *x,Integer *y,int cy) {
    unsigned long X,Y;
	if(cy) cy=1;
    for(int i=0 ; i<ALUSIZE ; i++) {
        X = x->n[i];
        Y = y->n[i];
        z->n[i] = X + Y + cy;
        cy = ((X>>1)+(Y>>1)+((((X&1)+(Y&1)+cy)>>1)&1))>>31;
    }
	z->carry = cy;
}
#else
void i_add(Integer *z,Integer *x,Integer *y,int cy) {
    unsigned long long X,Y;
	if(cy) cy=1;
    for(int i=0 ; i<ALUSIZE ; i++) {
        X = x->n[i];
        Y = y->n[i];
        z->n[i] = X + Y + cy;
        cy = ((X>>1)+(Y>>1)+((((X&1)+(Y&1)+cy)>>1)&1))>>63;
    }
	z->carry = cy;
}
#endif

#ifdef __EMSCRIPTEN__
void printon(Integer* x) {
	for(int i=ALUSIZE-1 ; i>=0 ; i--) printf("%08lX",x->n[i]);
}

void fprinton(Integer* x) {
	for(int i=ALUSIZE-1 ; i>=0 ; i--) fprintf(fresult,"%08lX",x->n[i]);
}
#else
void printon(Integer* x) {
	for(int i=ALUSIZE-1 ; i>=0 ; i--) printf("%016llX",x->n[i]);
}

void fprinton(Integer* x) {
	for(int i=ALUSIZE-1 ; i>=0 ; i--) fprintf(fresult,"%016llX",x->n[i]);
}
#endif

Integer A,B,C,D;
Integer M0A,M0B,M1A,M1B;
Integer R[16],ROLD;
int CF,ZF,DP,I,J,PC,PC0,PC1,RETADR,ECODE;
int X,Y;
unsigned int code;

void montomul();
void print_register(unsigned int);

inline void calc_adr(int inc) {
    inc &= 0xff;
    inc |= (0x80 & inc) << 1;
    PC = (PC + inc) & 0x1ff;
}

int iasm_main(const char*);

int main(int argc,char *argv[]) {
    Integer ADDL,ADDR,ADDZ,Z;
    char buf[1024];
    int PX,P[32];
    int CY,CMP,S0,S1,i;
    FILE *fp;

#ifdef __EMSCRIPTEN__
    if(sizeof(unsigned long)!=4) {
	    printf("sizeof(unsigned long) != 4\n");
		exit(-1);
	}
#else
    if(sizeof(unsigned long long)!=8) {
	    printf("sizeof(unsigned long long) != 8\n");
		exit(-1);
	}
#endif

	if (argc == 1) {
		iasm_main("infile");
	} else {
		iasm_main(argv[1]);
	}
	
    if((fp=fopen("bin","r"))==NULL) {
        printf("Can't open bin file\n");
        exit(-1);
    }
#ifdef __EMSCRIPTEN__
    if((fresult=fopen("output/result.txt","w"))==NULL) {
#else
    if((fresult=fopen("result.txt","w"))==NULL) {
#endif
        printf("Can't open result file\n");
        exit(-1);
    }
    for(i=0 ; i<16 ; i++) {
        fscanf(fp,"%s\n",buf);
        atoI(buf,&R[i]);
    }
    for(i=0 ; i<PLEN && !feof(fp) ; i++) fscanf(fp,"%8lx %8lx",&PCODE[i],&PDEBUG[i]);
	
    time = 0;
    PC1 = 0;
    PC0 = DP = 0;
    PC  = 1;
    do {
        code = PCODE[PC0];
        if(PDEBUG[PC0]) print_register(PDEBUG[PC0]);
        PC1 = PC0;
        PC0 = PC;
        PX = X;
        X = 0xf & (code >> 4);
        Y = 0xf &  code;
        for(i=0 ; i<32 ; i++) P[i]=0x1 & (code >> (31-i));
        if(P[13]) {
            i_copy(&ROLD,&R[Y]);
			i_copy(&R[Y],&A);
        }

        switch (code >> 28) {
        case  0: break;
        case  1: I=159; break;
        case  2: I=1023; break;
        case  3: I=code & 0x3ff; break;
        case  4: J=255; break;
        case  5: J=511; break;
        case  6: J=1023; break;
        case  7: PC = code & 0x1ff; break;
        case  8:
            if(ZF) calc_adr(code);
            break;
        case  9:
            if(testbit(&C,0) == 0)
                calc_adr(code);
            break;
        case 10:
            if(I!=0) calc_adr(code);
            I--;
            if(I<-1 || I>4096) {
                printf("Out of Range : I=%d\n",I);
                printf("PC = %d\n",PC);
                exit (-1);
            }
            break;
        case 11:
            if(J!=0) calc_adr(code);
            J--;
            if(J<-1 || J>4096) {
                printf("Out of Range : J=%d\n",J);
                printf("PC = %d\n",PC);
                exit (-1);
            }
            break;
        case 12:
            if(CF) calc_adr(code);
            break;
        case 13:
            RETADR = PC;
            calc_adr(code);
            break;
        case 14:
            PC = RETADR;
            break;
        case 15:
            ECODE = Y;
            break;
        }

        switch(P[4]<<1 | P[5]) {
        case 0: i_zero(&ADDL); break;
        case 1:
            if(P[12]==1 && testbit(&C,0)==0) i_zero(&ADDL);
            else i_copy(&ADDL,&A);
            break;
        case 2: 
		    i_copy(&ADDL,&A);
			i_lshift(&ADDL);
            ADDL.n[0] |= testbit(&B,LASTBIT);
            break;
        case 3: 
			printf("Undefine ADDL code\n");
            printf("PC = %d\n",PC);
            exit(-1);
        }

        switch(P[6]<<1 | P[7]) {
        case 0: i_zero(&ADDR); break;
        case 1: i_copy(&ADDR,&B); break;
        case 2: i_copy(&ADDR,&C); break;
        case 3: i_copy(&ADDR,&D); break;
        }

        if(P[9] == 1) for(int k=0 ; k <ALUSIZE ; k++) ADDR.n[k] = ~(ADDR.n[k]);

        if(P[22]  && P[23]) {
            if(P[12] == 0) {
			  i_copy(&ADDL,&M0A);
			  i_copy(&ADDR,&M0B);
            } else {
			  i_copy(&ADDL,&M1A);
			  i_copy(&ADDR,&M1B);
            }
        }

        if(P[8]) {
			int d = 0;
			for(int k=0 ; k<ALUSIZE ; k++) d |= D.n[k];
            if(d==0) ZF = 1;
            else ZF = 0;
        }
		
		i_add(&ADDZ,&ADDL,&ADDR,P[8]);
        CY = ADDZ.carry;

        if(P[22] && !P[23]) montomul();
        if(P[23] && !P[22]) {
		    i_zero(&M0A);
		    i_zero(&M0B);
		    i_zero(&M1A);
		    i_zero(&M1B);
            DP=0;
        }        

        if(P[11] == 1) {
            if(DP == 0) DP = 2;
		    else DP += 2;
            if(DP > 1025) DP -= 1024;
        }
		
        if(P[10]) i_copy(&Z,&R[PX]);
        else i_copy(&Z,&ADDZ);

        if(P[14]) CMP = testbit(&A,LASTBIT) | CY;
        else CMP = CF | CY;

        if(P[15]) CF = CY;

        /* D input select */
        switch (P[20]<<1 | P[21]) {
        case 0: break;
        case 1: i_copy(&D,&A); break;
        case 2: i_copy(&D,&Z); break;
        case 3: i_or(&D,&A,&Z); break;
        }

        /* A input select */
        S0 = P[14];
        if(P[11] && P[15]) S1 = CMP;
        else S1 = P[15];

        switch (S0<<1 | S1) {
        case 0: break;
        case 1: i_copy(&A,&Z); break;
        case 2: //A = A<<1 | testbit(B,1023);
		        i_lshift(&A);
				A.n[0] |= testbit(&B,LASTBIT);
                break;
        case 3: i_copy(&A,&Z); break;
        }

        /* B input select */
        switch (P[16]<<1 | P[17]) {
        case 0: break;
        case 1: 
		        i_copy(&B,&ADDZ);
				i_rshift(&B);
                if(CY == 1) setbit(&B,LASTBIT);
                break;
        case 2: i_copy(&B,&Z); break;
        case 3: //B = B << 1 | testbit(C,1023);
		        i_lshift(&B);
				B.n[0] |= testbit(&C,LASTBIT);
        }

        /* C input select */
        switch (P[18]<<1 | P[19]) {
        case 0: break;
        case 1: i_rshift(&C);
                if(testbit(&Z,0)) setbit(&C,LASTBIT);				
		        break;
        case 2: i_copy(&C,&Z); break;
        case 3: i_lshift(&C);
                C.n[0] |= CMP;
        }

        time ++;
        if( (code>>28) != 7) PC++;
    } while ((code >> 28) != 15 );

    printf("Sim end. end code = %d cyc = %d\n",ECODE,time);
    fprintf(fresult,"#Sim end. end code = %d cyc = %d\n",ECODE,time);
	fclose(fresult);
	return 0;
}

inline int csa5_s(int a,int b,int c,int d,int e) {
    return a^b^c^d^e;
}

inline int csa5_c(int a,int b,int c,int d,int e) {
    int x;
    x = a^b^c;
    return (x&d)|(x&e)|(d&e);
}
    
inline int csa5_w(int a,int b,int c,int d,int e) {
    return (a&b)|(a&c)|(b&c);
}

void montomul() {
    int i,j,u;
    int z;
    int f0,f1,f2,f3,f4;
    Integer A_I,B_I;
    int len;
    static int wA,wB;

    if(DP == 2) {
        i=0;
        wA = 0;
    } else {
        i = DP-2;
    }
    len = i+2;

    while(i<len) {
        z = testbit(&M0A,0) ^ testbit(&M0B,0);
        u = z ^ (testbit(&D,i) & testbit(&A,0));
		i_zero(&A_I);
		i_zero(&B_I);
        for(j=0 ; j<ALUBIT ; j++) {
            f0 = testbit(&M0A,j);
            f1 = testbit(&M0B,j);
            f2 = testbit(&A,j) & testbit(&D,i);
            f3 = u & testbit(&B,j);
            f4 = wA;
            
            if(csa5_c(f0,f1,f2,f3,f4)) setbit(&A_I,j);
            if(csa5_s(f0,f1,f2,f3,f4) && j!=0 ) setbit(&B_I,j-1);
            wA = csa5_w(f0,f1,f2,f3,f4);
        }
        if(wA == 1) setbit(&B_I,LASTBIT);
        i++;
		i_copy(&M0A,&A_I);
		i_copy(&M0B,&B_I);
    }

    if(DP == 2) {
        i=0;
        wB = 0;
    } else {
        i = DP-2;
    }
    len = i+2;

    while(i<len) {
        z = testbit(&M1A,0) ^ testbit(&M1B,0);
        u = z ^ (testbit(&D,i) & testbit(&D,0));
		i_zero(&A_I);
		i_zero(&B_I);
        for(j=0 ; j<ALUBIT ; j++) {
            f0 = testbit(&M1A,j);
            f1 = testbit(&M1B,j);
            f2 = testbit(&D,j) & testbit(&D,i);
            f3 = u & testbit(&B,j);
            f4 = wB;
            
            if(csa5_c(f0,f1,f2,f3,f4)) setbit(&A_I,j);
            if(csa5_s(f0,f1,f2,f3,f4) && j!=0 ) setbit(&B_I,j-1);
            wB = csa5_w(f0,f1,f2,f3,f4);
        }
        if(wB == 1) setbit(&B_I,LASTBIT);
        i++;
		i_copy(&M1A,&A_I);
		i_copy(&M1B,&B_I);
    }
}

void prt(const char*s,Integer* ip) {
    fprintf(fresult,"$%s",s);
    fprinton(ip);
	fprintf(fresult,"\n");
}

void print_register(unsigned int debug) {
    int i;
    unsigned int dt;
    unsigned long int OC;

	if(debug & 0x80000000) prt("A   =",&A);
    if(debug & 0x40000000) prt("B   =",&B);
    if(debug & 0x20000000) prt("C   =",&C);
    if(debug & 0x10000000) prt("D   =",&D);
    if(debug & 0x08000000) prt("M0A =",&M0A);
    if(debug & 0x04000000) prt("M0B =",&M0B);
    if(debug & 0x02000000) prt("M1A =",&M1A);
    if(debug & 0x01000000) prt("M1B =",&M1B);
    if(debug & 0x00800000) fprintf(fresult,"CF  =%5d\n",CF);
    if(debug & 0x00400000) fprintf(fresult,"ZF  =%5d\n",ZF);
    if(debug & 0x00200000) fprintf(fresult,"I   =%5d\n",I);
    if(debug & 0x00100000) fprintf(fresult,"J   =%5d\n",J);
    if(debug & 0x00080000) fprintf(fresult,"PC  =     %3X\n",PC);
    if(debug & 0x00040000) fprintf(fresult,"RETA=%5d\n",RETADR);
    if(debug & 0x00008000) prt("R0  = ",&R[0]);
    if(debug & 0x00004000) prt("R1  = ",&R[1]);
    if(debug & 0x00002000) prt("R2  = ",&R[2]);
    if(debug & 0x00001000) prt("R3  = ",&R[3]);
    if(debug & 0x00000800) prt("R4  = ",&R[4]);
    if(debug & 0x00000400) prt("R5  = ",&R[5]);
    if(debug & 0x00000200) prt("R6  = ",&R[6]);
    if(debug & 0x00000100) prt("R7  = ",&R[7]);
    if(debug & 0x00000080) prt("R8  = ",&R[8]);
    if(debug & 0x00000040) prt("R9  = ",&R[9]);
    if(debug & 0x00000020) prt("R10 = ",&R[10]);
    if(debug & 0x00000010) prt("R11 = ",&R[11]);
    if(debug & 0x00000008) prt("R12 = ",&R[12]);
    if(debug & 0x00000004) prt("R13 = ",&R[13]);
    if(debug & 0x00000002) prt("R14 = ",&R[14]);
    if(debug & 0x00000001) prt("R15 = ",&R[15]);
    if(debug & 0x00010000) {
        fprintf(fresult,"####################################################\n");
        for(i=0 ; i<16 ; i++) {
            fprinton(&R[i]);
			fprintf(fresult,"\n");
        }
        fprintf(fresult,"U    = %8X\n",code);
        fprintf(fresult,"CF   =         %d\n",CF);
        fprintf(fresult,"ZF   =         %d\n",ZF);
        fprintf(fresult,"I    =       %3X\n",I&0x3FF);
        fprintf(fresult,"J    =       %3X\n",J&0x3FF);
        fprintf(fresult,"RETA =       %3X\n",RETADR);
        fprintf(fresult,"PC0  =       %3X\n",PC0);
        fprintf(fresult,"PC   =       %3X\n",PC);
        prt("A    = ",&A);
        prt("B    = ",&B);
        prt("C    = ",&C);
        prt("D    = ",&D);
        fprintf(fresult,"PDRX =        %X\n",X);
        fprintf(fresult,"PDRY =        %X\n",Y);

        OC = PCODE[PC1];
        if((OC&0x00040000)&&((OC&0xf)==((OC&0xf0)>>4))) {
            prt("PDRD = ",&ROLD);
        } else {
            prt("PDRD = ",&R[X]);
        }        
        dt   = code >> 24;
        dt  &= 0xf;
        dt <<= 3;
        dt  |= (code & 0x00400000) >> 20;
        dt  |= (code & 0x00080000) >> 18;
        dt  |= ((code & 0x00000200) >> 9) &
               ((code & 0x00000100) >> 8);
        fprintf(fresult,"DT   =        %2X\n",dt);
    }
}
